bl_info = {
    "name": "FTP sync",
    "author": "Christophe Seux",
    "version": (1, 0),
    "blender": (2, 81, 0),
    "warning": "",
    "wiki_url": "",
    "category": "User",
}


import bpy
from bpy.types import Operator, AddonPreferences
from bpy.props import FloatVectorProperty
import os
from os.path import *
import re


### EXCLUSIONS
EXCLUSIONS = ['.*','*.db','*.blend1','*~','*sync-conflict*','*.DS_Store']


'''
TO DO
Accepts empty folders, option for type of sync (override etc)
'''

class Preferences(AddonPreferences):
    bl_idname = __name__

    domain : bpy.props.StringProperty(name="FTP Domain")
    login : bpy.props.StringProperty(name="FTP Login")
    password : bpy.props.StringProperty(name="FTP Password",subtype='PASSWORD')
    ftp_folder : bpy.props.StringProperty(name="FTP folder", default = '/')
    local_folder : bpy.props.StringProperty(name="Local Folder", subtype='DIR_PATH')
    active_connection : bpy.props.BoolProperty(name="Active Connection", default = True)
    sync_type : bpy.props.EnumProperty(
        name="Sync Type",
        items = [(i,i.title(),"") for i in ("ONLY_NEW", "OVERRIDE", "MORE_RECENT")],
        default = "ONLY_NEW")

    def draw(self, context):
        layout = self.layout
        layout.prop(self, "domain")
        layout.prop(self, "login")
        layout.prop(self, "password")
        layout.prop(self, "ftp_folder")
        layout.prop(self, "active_connection")
        layout.prop(self, "local_folder")



def connect_to_ftp(domain, login, passwd, active = True) :
    from ftplib import FTP
    ftp = FTP(domain,timeout = 10)
    ftp.login(user = login, passwd = passwd)
    ftp.set_pasv(active)

    return ftp


def is_exclude(name, patterns) :
    from fnmatch import fnmatch

    if not isinstance(patterns, (list,tuple)) :
        patterns = [patterns]

    return any([fnmatch(name, p) for p in patterns])


def get_ftp_files(ftp, root, exclusions) :
    files = []

    ftp.cwd(root)
    for k,v in ftp.mlsd() :
        if is_exclude(k, exclusions) : continue

        path = '/'.join([root,k])
        if v['type'] == 'file' :
            files.append(path)

        elif v['type'] == 'dir' :
            files+= get_ftp_files(ftp, path, exclusions)

    return sorted(files)

def get_files(root, exclusions) :
    '''Recursively get files in passed directory if not in exclusion list'''

    files = []
    for f in os.scandir(root) :
        if is_exclude(f.path, exclusions) : continue

        if f.is_file() :
            files.append(f.path)

        elif f.is_dir() :
            files+= get_files(f.path, exclusions)

    return sorted(files)


def createDirs(ftp, dirpath):
    from ftplib import error_perm

    """
    Create dir with subdirs (progressive dir creation).

    :param ftp:     connected FTP
    :param dirpath: path (like 'test/test1/test2')

    :type ftp:      FTP
    :type dirpath:  str
    :rtype:         None

    """

    dirpath = dirpath.replace('\\', '/')
    tmp = dirpath.split('/')
    dirs = []

    for _ in tmp:
        if len(dirs) == 0:
            dirs.append(_)
            continue

        dirs.append(dirs[-1] + '/' + _)

    for _ in dirs:
        try:
            ftp.mkd(_)
        except error_perm as e:
            e_str = str(e)
            if '550' in e_str and 'File exists' in e_str:
                continue


def exist_ftp_file(ftp, filepath) :
    filepath = filepath.replace('\\', '/')
    #ftp.cwd('/')

    if filepath.startswith('/') : filepath = filepath[1:]

    split_path = filepath.split('/')

    for i,path in enumerate(split_path) :
        list_dir = {k:v['type'] for k,v in ftp.mlsd()}
        #print(i,path, list(list_dir.keys()))
        if path in list_dir.keys() :
            if i == len(split_path)-1 :
                return True

            elif list_dir[path] == 'dir' :
                ftp.cwd(path)
            else :
                return False
        else :
            return False

class SendToFtp(Operator):
    """Send To FTP"""
    bl_idname = "ftpsync.send"
    bl_label = "Send to FTP"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        prefs = context.preferences.addons[__name__].preferences
        ftp = connect_to_ftp(prefs.domain, prefs.login, prefs.password, prefs.active_connection)

        print('\nStart Uploading to the FTP')
        print('Source Folder :', prefs.local_folder)

        ftp_folder = prefs.ftp_folder.replace('\\','/').strip('/')
        if not ftp_folder.startswith('/') : ftp_folder = '/'+ftp_folder

        local_folder = prefs.local_folder.replace('\\','/')


        for f in get_files(local_folder, EXCLUSIONS) :
            '''recursively get files in directory if not in exclusion list'''

            ftp.cwd(ftp_folder)
            dst_file = re.sub(local_folder, '', f.replace('\\','/')).strip('/')

            if prefs.sync_type == 'ONLY_NEW' and exist_ftp_file(ftp,dst_file) :
                continue

            ftp.cwd(ftp_folder)

            createDirs(ftp,dirname(dst_file))

            try :
                with open(f, 'rb') as fp:
                    ftp.storbinary('STOR '+dst_file, fp)

                print('New file :', dst_file)
            except :
                print('Impossible to copy %s'%fp)



        ftp.close()

        return {'FINISHED'}


class GetFromFtp(Operator,):
    """Create a new Mesh Object"""
    bl_idname = "ftpsync.get"
    bl_label = "Get from FTP"
    bl_options = {'REGISTER', 'UNDO'}


    def execute(self, context):
        prefs = context.preferences.addons[__name__].preferences
        ftp = connect_to_ftp(prefs.domain, prefs.login, prefs.password, prefs.active_connection)

        if not prefs.local_folder :
            print('You have to set the local folder in your prefs')
            return {'FINISHED'}

        print('\nStart Downloading from FTP')
        print('Destination Folder :', prefs.local_folder)

        ftp_folder = prefs.ftp_folder.replace('\\','/').strip('/')
        if not ftp_folder.startswith('/') : ftp_folder = '/'+ftp_folder

        local_folder = prefs.local_folder.replace('\\','/')

        for f in get_ftp_files(ftp, ftp_folder, EXCLUSIONS) :
            dst_file = join(local_folder, re.sub(ftp_folder,'',f.replace('\\','/')).strip('/'))

            if prefs.sync_type == 'ONLY_NEW' and exists(dst_file) :
                continue

            if not exists(dirname(dst_file)) :
                os.makedirs(dirname(dst_file))

            with open(dst_file, 'wb') as fp:
                ftp.retrbinary('RETR '+f, fp.write)

            print('New file :', dst_file)



        ftp.close()


        return {'FINISHED'}


class Menu(bpy.types.Menu):
    bl_label = "ADM"
    bl_idname = "ADM_MT_menu"

    def draw(self, context):
        layout = self.layout
        prefs = context.preferences.addons[__name__].preferences

        layout.operator("ftpsync.get", icon='IMPORT')
        layout.operator("ftpsync.send", icon='EXPORT')
        layout.prop(prefs, "sync_type", text = '')

def menu_draw(self,context):
    self.layout.menu("ADM_MT_menu")

cls = [Preferences, SendToFtp, GetFromFtp,Menu]



def register():
    bpy.types.TOPBAR_MT_editor_menus.append(menu_draw)
    for c in cls :
        bpy.utils.register_class(c)


def unregister():
    bpy.types.TOPBAR_MT_editor_menus.remove(menu_draw)
    for c in cls :
        bpy.utils.unregister_class(c)



if __name__ == "__main__":
    register()
